#define _WIN32_WINNT   0x0501
#include <windows.h>
#include <tchar.h>
#include <atlstr.h>
#include <atltime.h>
#include <psapi.h>
#include <sddl.h>

#pragma comment(lib, "advapi32")
#pragma comment(lib, "psapi")

#define SERVICE_NAME _T("MyServiceX")
#define ACCEPTED_CONTROLS   (SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_SHUTDOWN)
#define SERVICE_CONTROL_CUSTOM 128
#define COUNTOF(x)  (sizeof(x)/sizeof(x[0]))

SERVICE_STATUS status = {0};
SERVICE_STATUS_HANDLE hStatus = NULL;

struct CLogger
{
	FILE* f;
	CLogger();
	~CLogger();
	void PutStr(LPCTSTR str);
} logger;

CLogger::CLogger()
{
	TCHAR path[MAX_PATH] = {0};
	GetModuleFileName(GetModuleHandle(NULL), path, MAX_PATH);
	LPTSTR p = _tcsrchr(path, _T('\\'));
	if (p)
	{
		_tcscpy(p, _T("\\mysvcx.log"));
		f = _tfopen(path, _T("a+"));
		PutStr(_T("++ Start ++"));
	}
}

CLogger::~CLogger()
{
	PutStr(_T("-- Stop --"));
	fclose (f);
}

void CLogger::PutStr(LPCTSTR str)
{
	if (!f)
		return;

	_ftprintf(f, _T("[%s] %s\n"), CTime::GetCurrentTime().Format(_T("%c")), str);
	fflush(f);
}

DWORD GetProcessTokenByNameAndUser(LPCTSTR procName, LPCTSTR user, HANDLE* phToken)
{
	typedef BOOL (WINAPI* PFNQueryFullProcessImageName)(HANDLE, DWORD, LPTSTR, PDWORD);
	PFNQueryFullProcessImageName pfnQueryFullProcessImageName =
		(PFNQueryFullProcessImageName)::GetProcAddress(::GetModuleHandle(_T("kernel32.dll")), "QueryFullProcessImageNameW");

	DWORD procs[1024] = {0};
	DWORD cb = 0, pid = 0;
	if (!EnumProcesses(procs, sizeof(procs), &cb))
		return 0;

	for (int idx = 0; idx < cb/sizeof(DWORD); idx++)
	{
		pid = procs[idx];
		TCHAR path[MAX_PATH] = {0};
		DWORD cch = 0;

		HANDLE hToken = NULL;
		DWORD procAccess = PROCESS_ALL_ACCESS;
		HANDLE hProc = OpenProcess(procAccess, FALSE, pid);
		if (NULL != hProc)
		{
			TCHAR userName[48] = {0}, domainName[48] = {0};

			BOOL bSuccess = GetModuleFileNameEx(hProc, NULL, path, MAX_PATH);

			if (!bSuccess)
			{
#if 0
				DWORD err = GetLastError();
				CString msg;
				msg.Format(_T("[%d] GetModuleFileNameEx err: %d (%08d)"), pid, err, err);
				logger.PutStr(msg);
#endif
				if (pfnQueryFullProcessImageName)
				{
					DWORD cch = COUNTOF(path);
					bSuccess = pfnQueryFullProcessImageName(hProc, 0, path, &cch);
				}

			}

			if (bSuccess)
			{
				CString msg;
				msg.Format(_T("[PID:%d] Path: %s"), pid, path);
				LPTSTR p = _tcsrchr(path, _T('\\'));
				if (CString(p + 1).CompareNoCase(procName))
				{
					CloseHandle(hProc);
					continue;
				}
				logger.PutStr(msg);
				if (OpenProcessToken(hProc, TOKEN_ALL_ACCESS, &hToken))
				{
					msg.Format(_T("[%d] Token: %p"), pid, hToken);
					logger.PutStr(msg);
					TOKEN_USER* ptu = NULL;
					DWORD size = 0;

					GetTokenInformation(hToken, TokenUser, NULL, 0, &size);
					if (size)
					{
						ptu = (TOKEN_USER*)LocalAlloc(LPTR, size);
						if (ptu && GetTokenInformation(hToken, TokenUser, ptu, size, &size))
						{
							LPTSTR sidStr = NULL;
							if (ConvertSidToStringSid(ptu->User.Sid, &sidStr))
							{
								msg.Format(_T("[%d] uSID: %s"), pid, sidStr);
								logger.PutStr(msg);
							}
							LocalFree(sidStr);

							DWORD cchUser = COUNTOF(userName), cchDomain = COUNTOF(domainName);
							SID_NAME_USE use;
							if (LookupAccountSid(NULL, ptu->User.Sid, 
								userName, &cchUser, 
								domainName, &cchDomain,
								&use))
							{
								msg.Format(_T("[%d] User: %s\\%s"), pid, domainName, userName);
								logger.PutStr(msg);
							}
							else
							{
								msg.Format(_T("[%d] LookupAccountSid failed"), pid);
								logger.PutStr(msg);
							}
							LocalFree(ptu);
						}
						else
						{
							msg.Format(_T("[%d] GetTokenInformation failed"), pid);
							logger.PutStr(msg);
						}
					}
					else
					{
						msg.Format(_T("[%d] Zero token info size returned"), pid);
						logger.PutStr(msg);
					}
				}
#if 0
				{
					DWORD err = GetLastError();
					msg.Format(_T("[%d] Last Error: %d"), pid, err);
					logger.PutStr(msg);
				}
#endif

				CString userAcc = CString(domainName) + _T("\\") + CString(userName);
				if (0 == userAcc.CompareNoCase(user))
				{
					if (phToken)
						*phToken = hToken;
					break;
				}
			}
		}

		CloseHandle(hProc);
	}

	return pid;
}

DWORD CALLBACK ExecThread(LPVOID pVoid)
{
	HANDLE hToken = NULL;
	STARTUPINFO si = {sizeof(si)};
	PROCESS_INFORMATION pi = {0};
	si.lpDesktop = _T("WinSta0\\Default");
	TCHAR path[MAX_PATH] = {0};
	CString msg;

	logger.PutStr(_T("Duplicating token handle..."));

	BOOL ret = DuplicateHandle(
		GetCurrentProcess(),
		(HANDLE)pVoid,
		GetCurrentProcess(),
		&hToken,
		0,
		FALSE,
		DUPLICATE_CLOSE_SOURCE | DUPLICATE_SAME_ACCESS);

	if (ret)
	{
		logger.PutStr(_T("Impersonating logged on user..."));
		ret = ImpersonateLoggedOnUser(hToken);
	}
	else
	{
		DWORD err = GetLastError();
		msg.Format(_T("DuplicateHandle failed (%d)"), err);
		logger.PutStr(msg);
	}

	if (ret)
	{
		ret = ExpandEnvironmentStrings(_T("%SystemRoot%\\system32\\calc.exe"), path, MAX_PATH);
	}
	else
	{
		DWORD err = GetLastError();
		msg.Format(_T("ImpersonateLoggedOnUser failed (%d)"), err);
		logger.PutStr(msg);
	}

	if (ret)
	{
		logger.PutStr(_T("Running executable..."));
		logger.PutStr(path);
		ret = CreateProcessAsUser(hToken,
			NULL, path, NULL, NULL, 
			FALSE, 0, NULL, NULL, &si, &pi);
	}
	else
	{
		DWORD err = GetLastError();
		msg.Format(_T("ExpandEnvironmentStrings failed (%d)"), err);
		logger.PutStr(msg);
	}

	if (ret)
	{
		msg.Format(_T("[PID:%d] Process was run successfully. Waiting for stop..."), pi.dwProcessId);
		logger.PutStr(msg);
		WaitForSingleObject(pi.hProcess, INFINITE);
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
		logger.PutStr(_T("Cleanup after process stop..."));
	}
	else
	{
		DWORD err = GetLastError();
		msg.Format(_T("CreateProcessAsUser failed (%d)"), err);
		logger.PutStr(msg);
	}

	RevertToSelf();

	logger.PutStr(_T("Quitting thread..."));

	return 0;
}

void RunUserApp()
{
	HANDLE hToken = NULL;
	DWORD pid = GetProcessTokenByNameAndUser(_T("explorer.exe"), _T("IVARTANOV-W7\\IGOR"), &hToken);
	CString msg;
	msg.Format(_T("Token: %08x, pid: %d"), hToken, pid);
	logger.PutStr(msg);
	DWORD threadId = 0;
	HANDLE hThread = CreateThread(NULL, 0, ExecThread, hToken, 0, &threadId);
	WaitForSingleObject(hThread, 2000);
	CloseHandle(hThread);
}

CString GetExeFolderName()
{
	TCHAR path[MAX_PATH] = {0};
	GetModuleFileName(GetModuleHandle(NULL), path, MAX_PATH);
	LPTSTR p = _tcsrchr(path, _T('\\'));
	if (p)
	{
		*p = TEXT('\0');
		return path;
	}
	return (LPCTSTR)NULL;
}

DWORD CALLBACK ExeThread(LPVOID pVoid)
{
    CString path = TEXT("C:\\testService\\Debug\\testService.exe");
    logger.PutStr(path);

	STARTUPINFO info = {0};
	PROCESS_INFORMATION processInfo;

	ZeroMemory( &info, sizeof(info) );
	info.cb = sizeof(info);
	ZeroMemory( &processInfo, sizeof(processInfo) );

	info.dwFlags = STARTF_USESHOWWINDOW;
	info.wShowWindow = FALSE; 

	if (CreateProcess(path, NULL, NULL, NULL, TRUE, 0, NULL, NULL, &info, &processInfo))
	{
		logger.PutStr(TEXT("Waiting for test.exe stop..."));
		::WaitForSingleObject(processInfo.hProcess, INFINITE);
		CloseHandle(processInfo.hProcess);
		CloseHandle(processInfo.hThread);
		logger.PutStr(TEXT("Detected test.exe stop."));
    }

    logger.PutStr(TEXT("Quitting thread."));

    return 0;
}

void RunApp()
{
	DWORD threadId = 0;
	HANDLE hThread = CreateThread(NULL, 0, ExeThread, NULL, 0, &threadId);
	Sleep(20);
	CloseHandle(hThread);
}

#include "test.h"

DWORD WINAPI HandlerEx(DWORD dwControl, DWORD dwEventType, LPVOID lpEventData, LPVOID lpContext)
{
	DWORD retCode = NO_ERROR;

	switch(dwControl)
	{
	case SERVICE_CONTROL_STOP:
	case SERVICE_CONTROL_SHUTDOWN:
		status.dwCurrentState = SERVICE_STOPPED;
		break;
	case SERVICE_CONTROL_CUSTOM:
		RunUserApp();
		status.dwCurrentState = SERVICE_RUNNING;
		break;
	case SERVICE_CONTROL_CUSTOM + 1:
		RunApp();
		status.dwCurrentState = SERVICE_RUNNING;
		break;
	case SERVICE_CONTROL_CUSTOM + 2:
		{
			HANDLE hEvent = CreateEvent(NULL, TRUE, FALSE, EVENT_NAME);
			SetEvent(hEvent);
			Sleep(20);
			CloseHandle(hEvent);
		}
		break;
	default:
		retCode = ERROR_CALL_NOT_IMPLEMENTED;
	}

	SetServiceStatus(hStatus, &status);
	return retCode;
}

VOID WINAPI ServiceMain(DWORD dwArgc, LPTSTR* lpszArgv)
{
	status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	status.dwControlsAccepted = ACCEPTED_CONTROLS;
	status.dwCurrentState = SERVICE_START_PENDING;

	hStatus = RegisterServiceCtrlHandlerEx(SERVICE_NAME, HandlerEx, NULL);
	status.dwCurrentState = (NULL != hStatus) ? SERVICE_RUNNING : SERVICE_STOPPED;
	SetServiceStatus(hStatus, &status);
}

int _tmain()
{
	SERVICE_TABLE_ENTRY st[] = { {SERVICE_NAME, ServiceMain}, {NULL, NULL} };
	BOOL ret = StartServiceCtrlDispatcher(st);
	return 0;
}
